import asyncio

from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from fastapi.responses import HTMLResponse

from models import MusicPlayer, PlayerState, WSConnectionType
from services import ConnectionManager, DownloadService

# Setup
tags_metadata = [
    {"name": "player", "description": "Interact with the Music Player"},
    {"name": "experimental"},
    {"name": "queue"},
]
app = FastAPI(openapi_tags=tags_metadata)
player = MusicPlayer()
dl_service = DownloadService()
ws_manager = ConnectionManager()


# Interface
@app.get("/", response_class=HTMLResponse)
async def root():
    with open("index.html") as f:
        return f.read()


@app.on_event("startup")
async def start_event_loop():
    asyncio.create_task(state_broadcast_loop())
    asyncio.create_task(queue_broadcast_loop())


async def state_broadcast_loop():
    while True:
        await player._state_event.wait()
        await ws_manager.broadcast(WSConnectionType.state, player.get_state())
        player._state_event.clear()


async def queue_broadcast_loop():
    while True:
        await player._queue_event.wait()
        await ws_manager.broadcast(WSConnectionType.queue, player.get_queue())
        player._queue_event.clear()


# Status updates
@app.websocket("/player")
async def websocket_player(websocket: WebSocket):
    await ws_manager.connect(websocket, WSConnectionType.state)
    try:
        while True:
            await websocket.receive_text()
            await ws_manager.send(websocket, player.get_state())
    except WebSocketDisconnect:
        ws_manager.disconnect(websocket)


# Queue updates
@app.websocket("/queue")
async def websocket_queue(websocket: WebSocket):
    await ws_manager.connect(websocket, WSConnectionType.queue)
    try:
        while True:
            await websocket.receive_text()
            await ws_manager.send(websocket, player.get_queue())
    except WebSocketDisconnect:
        ws_manager.disconnect(websocket)


@app.get("/queue", tags=["queue"])
def get_queue():
    return player.get_queue()


@app.post("/queue", tags=["queue"])
async def post_to_queue(url: str):
    track = await dl_service.download(url)
    await player.add_to_queue(track)


@app.delete("/queue/{track_id}", tags=["queue"])
def delete_from_queue(track_id: str):
    player.remove_from_queue_by_id(track_id)


@app.post("/player/play", tags=["player"])
async def player_play():
    await player.play()


@app.post("/player/stop", tags=["player"])
async def player_stop():
    await player.stop()


@app.post("/player/skip", tags=["player"])
async def player_skip():
    await player.next()


# Player
@app.put("/player/volume", tags=["player"])
async def set_volume(volume: float):
    await player.set_volume(volume)


@app.get("/player/volume", tags=["player"])
def get_volume():
    return {"volume": player.get_volume()}


@app.get("/player", tags=["player"])
def get_player_state() -> PlayerState:
    return player.get_state()
